# © Copyright Databand.ai, an IBM Company 2022

import logging

from datetime import timedelta

from dbnd_examples_orchestration.data import data_repo
from dbnd_examples_orchestration.orchestration.features.partitioned_calculations_pipeline import (
    ExamplePartitionedCalculations,
)
from dbnd_examples_orchestration.orchestration.features.partitioned_data_pipeline import (
    ExamplePartitionedDataPipeline,
)
from dbnd_examples_orchestration.orchestration.features.partitioned_inputs import (
    DeviceLogsPipeline,
    ExamplePartitionedPipeline,
)

from dbnd_run.testing.helpers import assert_run_task


logger = logging.getLogger(__name__)


class TestFeaturePartitions(object):
    def test_partitions(self):
        target = ExamplePartitionedPipeline(
            task_target_date=data_repo.partitioned_data_target_date
        )
        assert_run_task(target)

    def test_pipeline_with_logs(self):
        task = assert_run_task(
            ExamplePartitionedPipeline(
                task_target_date=data_repo.partitioned_data_target_date
            )
        )
        logger.error(task.features.read())

    def test_paritioned_input(self):
        assert_run_task(
            DeviceLogsPipeline(
                period=timedelta(days=1),
                task_target_date=data_repo.partitioned_data_target_date,
            )
        )

    def test_partitioned_data_task(self):
        task = ExamplePartitionedCalculations()
        assert_run_task(task)

    def test_partitioned_batch_task(self):
        task = ExamplePartitionedDataPipeline()
        assert_run_task(task)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.orchestration.customizations.custom_decorator import (
    my_new_experiement,
)


class TestCustomDecorator(object):
    def test_my_new_experiement(self):
        task = my_new_experiement.dbnd_run().root_task
        assert task



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd_examples_orchestration.orchestration.customizations.custom_target import (
    CustomIOPipeline,
)

from dbnd_run.testing.helpers import assert_run_task


logger = logging.getLogger(__name__)


class TestRunWithCustomTargets(object):
    def test_parameter_with_multiple_partitions(self):
        assert_run_task(CustomIOPipeline())



# © Copyright Databand.ai, an IBM Company 2022

import pandas as pd
import tensorflow as tf


def train_input_fn(features, labels, batch_size):
    # Convert the inputs to a Dataset.
    dataset = tf.data.Dataset.from_tensor_slices((dict(features), labels))

    # Shuffle, repeat, and batch the dbnd_examples_orchestration.
    dataset = dataset.shuffle(1000).repeat().batch(batch_size)

    # Return the dataset.
    return dataset


def eval_input_fn(features, labels, batch_size):
    features = dict(features)
    if labels is None:
        # No labels, use only features.
        inputs = features
    else:
        inputs = (features, labels)

    # Convert the inputs to a Dataset.
    dataset = tf.data.Dataset.from_tensor_slices(inputs)

    # Batch the dbnd_examples_orchestration
    dataset = dataset.batch(batch_size)
    return dataset


def build_classifier(train, model_dir):
    my_feature_columns = []
    for key in train.keys():
        my_feature_columns.append(tf.feature_column.numeric_column(key=key))

    # Build 2 hidden layer DNN with 10, 10 units respectively.
    classifier = tf.estimator.DNNClassifier(
        feature_columns=my_feature_columns,
        hidden_units=[10, 10],
        n_classes=3,
        optimizer=tf.compat.v1.train.ProximalAdagradOptimizer(
            learning_rate=0.001, l1_regularization_strength=0.01
        ),
        model_dir=model_dir,
        config=tf.estimator.RunConfig().replace(save_summary_steps=10),
    )
    return classifier


def create_receiver_fn():
    feature_spec = {
        "SepalLength": tf.compat.v1.placeholder(
            "float", name="SepalLength", shape=[None]
        ),
        "SepalWidth": tf.compat.v1.placeholder(
            "float", name="SepalWidth", shape=[None]
        ),
        "PetalLength": tf.compat.v1.placeholder(
            "float", name="PetalLength", shape=[None]
        ),
        "PetalWidth": tf.compat.v1.placeholder(
            "float", name="PetalWidth", shape=[None]
        ),
    }
    receiver_fn = tf.estimator.export.build_raw_serving_input_receiver_fn(feature_spec)
    return receiver_fn


def predict(data, model_path):
    graph = tf.Graph()
    with tf.compat.v1.Session(graph=graph) as sess:
        meta_graph_def = tf.compat.v1.saved_model.loader.load(
            sess, [tf.compat.v1.saved_model.tag_constants.SERVING], model_path
        )
        model_signature = meta_graph_def.signature_def["predict"].inputs

        fetch_mapping = {
            sigdef_output: graph.get_tensor_by_name(tnsr_info.name)
            for sigdef_output, tnsr_info in model_signature.outputs.items()
        }

        inputs = {
            graph.get_tensor_by_name(tnsr_info.name): data[sigdef_input]
            for sigdef_input, tnsr_info in model_signature.inputs.items()
        }

        raw_predictions = sess.run(fetch_mapping, feed_dict=inputs)
        predictions_as_dict = {
            fetch_name: list(values) for fetch_name, values in raw_predictions.items()
        }
        return pd.DataFrame(data=predictions_as_dict)



# © Copyright Databand.ai, an IBM Company 2022

import pandas as pd

from dbnd_examples_orchestration.data import data_repo

from dbnd import pipeline, task


@task
def prepare_data(alpha=0.5):
    return alpha


class TestDocTaskObject:
    def test_task_class_version(self):
        #### DOC START
        @task(task_class_version=2)
        def prepare_data(data: pd.DataFrame) -> pd.DataFrame:
            return data

        #### DOC END
        prepare_data.dbnd_run(data=data_repo.wines)

    def test_doc(self):
        #### DOC START
        c = prepare_data(alpha=0.5)
        d = prepare_data(alpha=0.5)

        assert c is d
        #### DOC END

    def test_prepare_data_pipeline(self):
        #### DOC START
        @pipeline
        def prepare_data_pipeline():
            prepare_data(task_name="first_prepare_data")
            prepare_data(task_name="second_prepare_data")

        #### DOC END
        prepare_data_pipeline.dbnd_run()

    def test_task_in_task(self):
        #### DOC START
        import logging

        from dbnd import task

        @task
        def calculate_alpha(alpha: int):
            calculated_alpha = "alpha: {}".format(alpha)
            logging.info(calculated_alpha)
            return calculated_alpha

        @task
        def prepare_data(data, additional_data):
            return "{} {}".format(data, additional_data)

        @task
        def prepare_dataset(data, data_num: int = 3) -> str:
            result = ""
            for i in range(data_num):
                result = prepare_data(result, data)

            return result

        #### DOCK END
        prepare_dataset.dbnd_run(data=data_repo.wines)
        calculate_alpha.dbnd_run(alpha=1)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Dict, List

from pandas import DataFrame

from dbnd import band, log_dataframe, parameter, task
from dbnd.utils import data_combine
from targets import Target


logger = logging.getLogger(__name__)


@task
def dump_table_from_db(table):
    # type: (str) -> List[str]
    return ["raw_id,partner,value\n"] + [
        "%s_%s,%s,%s\n" % (table, idx, idx % 3, idx * 2) for idx in range(10)
    ]


@task
def filter_by_id(data, partner_id):
    # type: (DataFrame, int) -> DataFrame
    assert isinstance(data, DataFrame)
    partner_data = data[data["partner"] == partner_id]
    log_dataframe("partner_%s" % partner_id, partner_data)
    return partner_data


@task(partners_files=parameter.data)
def build_report_one_by_one(partners_files):
    # type: (Dict[int,Target]) -> str
    logger.info("partners files: %s", partners_files)
    return "OK"


@task
def build_report_from_dataframe(partners_df):
    # type: (DataFrame) -> str
    logger.info("partners data: %s", partners_df)
    return str(len(partners_df))


@band
def output_per_id_report(partners):
    # type: (List[int]) -> ...
    dump = dump_table_from_db(table="my_table")
    partners_info = {partner: filter_by_id(dump, partner) for partner in partners}

    report_one_by_one = build_report_one_by_one(partners_files=partners_info)

    report_from_dataframe = build_report_from_dataframe(
        partners_df=data_combine(partners_info.values())
    )
    return report_one_by_one, report_from_dataframe



# © Copyright Databand.ai, an IBM Company 2022

import datetime

import numpy as np
import pandas as pd

from dbnd_examples_orchestration.data import demo_data_repo, partner_data_file

from dbnd import output, pipeline, task
from dbnd.utils import period_dates


@pipeline
def generate_partner_data(
    seed: pd.DataFrame = demo_data_repo.seed,
    task_target_date=datetime.datetime.now().date(),
    period=datetime.timedelta(days=7),
    bad_labels_date=datetime.datetime.strptime("2018-01-01", "%Y-%m-%d").date(),
):
    data = rename_columns(data=seed)
    data = calculate_target_variable(data=data)

    results = {}

    for d in period_dates(task_target_date, period):
        r = generate_for_date(
            task_target_date=d,
            data=data,
            a_out=demo_data_repo.partner_a_file(d),
            b_out=demo_data_repo.partner_b_file(d),
            c_out=demo_data_repo.partner_c_file(d),
        )
        results[str(d)] = r

    noisy_data = generate_for_date(
        task_target_date=bad_labels_date,
        data=data,
        noise=True,
        a_out=demo_data_repo.partner_a_file(bad_labels_date),
        b_out=demo_data_repo.partner_b_file(bad_labels_date),
        c_out=demo_data_repo.partner_c_file(bad_labels_date),
    )
    results[str(bad_labels_date)] = noisy_data

    customers = create_customer_files(
        seed=data,
        a_out=partner_data_file("customer_a.csv"),
        b_out=partner_data_file("customer_b.csv"),
    )
    results["customers"] = customers
    return results


@task
def create_customer_files(
    seed: pd.DataFrame, a_out=output.csv[pd.DataFrame], b_out=output.csv[pd.DataFrame]
):
    data = seed.copy(deep=True)
    data.to_target(a_out, index=False)
    data["target"] = np.random.randint(
        low=data.shape[0], high=100 * data.shape[0], size=data.shape[0]
    )
    data.to_target(b_out, index=False)
    return "Ok"


@task
def rename_columns(data: pd.DataFrame) -> pd.DataFrame:
    data.columns = list("0123456789") + ["10", "target"]
    return data


@task
def calculate_target_variable(data: pd.DataFrame) -> pd.DataFrame:
    data = data.drop(["target"], axis=1)
    ids = np.arange(data.shape[1])
    coefficients = (-1) ** ids * np.exp(-ids / 10)
    y = np.dot(data.values, coefficients)
    y += 0.01 * np.random.normal(size=data.shape[0])
    data["target"] = y

    return data


@task
def generate_for_date(
    task_target_date,
    data: pd.DataFrame,
    a_out=output.csv[pd.DataFrame],
    b_out=output.json[pd.DataFrame],
    c_out=output.csv[pd.DataFrame],
    noise=False,
):
    sample = data.sample(frac=0.2, random_state=task_target_date.day)
    partner_a = sample[list("012345")].copy()
    partner_b = sample[list("6789")].copy()
    partner_c = sample[["10", "target"]].copy()
    if noise:
        partner_c["target"] = np.random.randint(
            low=partner_c.shape[0],
            high=100 * partner_c.shape[0],
            size=partner_c.shape[0],
        )

    partner_a.index.names = ["id"]
    partner_b.index.names = ["id"]
    partner_b.reset_index(inplace=True)
    partner_c.index.names = ["id"]

    partner_a.to_target(a_out)
    partner_b.to_target(b_out)
    partner_c.to_target(c_out)

    return "Ok"


if __name__ == "__main__":
    generate_partner_data.dbnd_run(seed=demo_data_repo.seed, task_version="now")



# © Copyright Databand.ai, an IBM Company 2022

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by databand.ai to for more advance example.

"""A word-counting workflow."""

from __future__ import absolute_import

import argparse
import logging
import re

import apache_beam as beam

from apache_beam.io import ReadFromText, WriteToText
from apache_beam.metrics import Metrics
from apache_beam.metrics.metric import MetricsFilter
from apache_beam.options.pipeline_options import PipelineOptions, SetupOptions
from past.builtins import unicode


class WordExtractingDoFn(beam.DoFn):
    """Parse each line of input text into words."""

    def __init__(self):
        self.words_counter = Metrics.counter(self.__class__, "words")
        self.word_lengths_counter = Metrics.counter(self.__class__, "word_lengths")
        self.word_lengths_dist = Metrics.distribution(self.__class__, "word_len_dist")
        self.empty_line_counter = Metrics.counter(self.__class__, "empty_lines")

    def process(self, element):
        """Returns an iterator over the words of this element.

        The element is a line of text.  If the line is blank, note that, too.

        Args:
          element: the element being processed

        Returns:
          The processed element.
        """
        text_line = element.strip()
        if not text_line:
            self.empty_line_counter.inc(1)
        words = re.findall(r"[\w\']+", text_line, re.UNICODE)
        for w in words:
            self.words_counter.inc()
            self.word_lengths_counter.inc(len(w))
            self.word_lengths_dist.update(len(w))
        return words


def run(argv=None):
    """Main entry point; defines and runs the wordcount pipeline."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input",
        dest="input",
        default="gs://dataflow-samples/shakespeare/kinglear.txt",
        help="Input file to process.",
    )
    parser.add_argument(
        "--output",
        dest="output",
        required=True,
        help="Output file to write results to.",
    )
    known_args, pipeline_args = parser.parse_known_args(argv)

    # We use the save_main_session option because one or more DoFn's in this
    # workflow rely on global context (e.g., a module imported at module level).
    pipeline_options = PipelineOptions(pipeline_args)
    pipeline_options.view_as(SetupOptions).save_main_session = True
    p = beam.Pipeline(options=pipeline_options)

    # Read the text file[pattern] into a PCollection.
    lines = p | "read" >> ReadFromText(known_args.input)

    # Count the occurrences of each word.
    def count_ones(word_ones):
        (word, ones) = word_ones
        return (word, sum(ones))

    counts = (
        lines
        | "split" >> (beam.ParDo(WordExtractingDoFn()).with_output_types(unicode))
        | "pair_with_one" >> beam.Map(lambda x: (x, 1))
        | "group" >> beam.GroupByKey()
        | "count" >> beam.Map(count_ones)
    )

    # Format the counts into a PCollection of strings.
    def format_result(word_count):
        (word, count) = word_count
        return "%s: %d" % (word, count)

    output = counts | "format" >> beam.Map(format_result)

    # Write the output using a "Write" transform that has side effects.
    # pylint: disable=expression-not-assigned
    output | "write" >> WriteToText(known_args.output)

    result = p.run()
    result.wait_until_finish()

    # Do not query metrics when creating a template which doesn't run
    if (
        not hasattr(result, "has_job") or result.has_job  # direct runner
    ):  # not just a template creation
        empty_lines_filter = MetricsFilter().with_name("empty_lines")
        query_result = result.metrics().query(empty_lines_filter)
        if query_result["counters"]:
            empty_lines_counter = query_result["counters"][0]
            logging.info("number of empty lines: %d", empty_lines_counter.result)

        word_lengths_filter = MetricsFilter().with_name("word_len_dist")
        query_result = result.metrics().query(word_lengths_filter)
        if query_result["distributions"]:
            word_lengths_dist = query_result["distributions"][0]
            logging.info("average word length: %d", word_lengths_dist.result.mean)


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)
    run()



# © Copyright Databand.ai, an IBM Company 2022

import mxnet as mx

from dbnd_examples_orchestration.data.tool_mxnet import digit_data
from dbnd_examples_orchestration.src.tool_mxnet.mxnet_task import (
    DownloadFile,
    MXNetTask,
)

from dbnd import PipelineTask, data, namespace, output, parameter


namespace("digits")


def create_mlp_digit_classifier():
    data = mx.sym.var("data")
    # Flatten the data from 4-D shape into 2-D (batch_size, num_channel*width*height)
    data = mx.sym.flatten(data=data)
    # The first fully-connected layer and the corresponding activation function
    fc1 = mx.sym.FullyConnected(data=data, num_hidden=128)
    act1 = mx.sym.Activation(data=fc1, act_type="relu")

    # The second fully-connected layer and the corresponding activation function
    fc2 = mx.sym.FullyConnected(data=act1, num_hidden=64)
    act2 = mx.sym.Activation(data=fc2, act_type="relu")

    # MNIST has 10 classes
    fc3 = mx.sym.FullyConnected(data=act2, num_hidden=10)
    # Softmax with cross entropy loss
    mlp = mx.sym.SoftmaxOutput(data=fc3, name="softmax")
    return mlp


def create_lenet_digit_classifier():
    data = mx.sym.var("data")
    # first conv layer
    conv1 = mx.sym.Convolution(data=data, kernel=(5, 5), num_filter=20)
    tanh1 = mx.sym.Activation(data=conv1, act_type="tanh")
    pool1 = mx.sym.Pooling(data=tanh1, pool_type="max", kernel=(2, 2), stride=(2, 2))
    # second conv layer
    conv2 = mx.sym.Convolution(data=pool1, kernel=(5, 5), num_filter=50)
    tanh2 = mx.sym.Activation(data=conv2, act_type="tanh")
    pool2 = mx.sym.Pooling(data=tanh2, pool_type="max", kernel=(2, 2), stride=(2, 2))
    # first fullc layer
    flatten = mx.sym.flatten(data=pool2)
    fc1 = mx.symbol.FullyConnected(data=flatten, num_hidden=500)
    tanh3 = mx.sym.Activation(data=fc1, act_type="tanh")
    # second fullc
    fc2 = mx.sym.FullyConnected(data=tanh3, num_hidden=10)
    # softmax loss
    lenet = mx.sym.SoftmaxOutput(data=fc2, name="softmax")
    return lenet


class TrainModel(MXNetTask):
    learning_rate = parameter.default(0.1)[float]
    optimizer = parameter.default("sgd")
    epoch_num = parameter[int].default(10)

    training_data = data
    training_labels = data

    test_data = data
    test_labels = data

    model = output

    def main(self, ctx):
        train_iter = self.to_ndarray_iterator(
            self.training_data, self.training_labels, shuffle=True
        )
        val_iter = self.to_ndarray_iterator(self.test_data, self.test_labels)

        # create a trainable module on compute context
        mlp_model = mx.mod.Module(symbol=create_mlp_digit_classifier(), context=ctx)
        mlp_model.fit(
            train_iter,  # train data
            eval_data=val_iter,  # validation data
            optimizer=self.optimizer,  # use SGD to train
            optimizer_params={
                "learning_rate": self.learning_rate
            },  # use fixed learning rate
            eval_metric="acc",  # report accuracy during training
            batch_end_callback=mx.callback.Speedometer(self.batch_size),
            # output progress for each 100 data batches
            num_epoch=self.epoch_num,
        )  # train for at most 10 dataset passes

        # mlp_model.save_checkpoint("mnist", self.epoch_num)
        mlp_model.save_params(self.model.path)


class ValidateModel(MXNetTask):
    model = data
    validation_data = data
    validation_label = data

    report = output

    def main(self, ctx):
        test_iter = self.to_ndarray_iterator(
            self.validation_data, self.validation_label
        )

        loaded = mx.mod.Module(symbol=create_mlp_digit_classifier(), context=ctx)
        loaded.bind(
            data_shapes=test_iter.provide_data, label_shapes=test_iter.provide_label
        )
        loaded.load_params(self.model.path)

        acc = mx.metric.Accuracy()
        mae = mx.metric.MAE()

        loaded.score(test_iter, [acc, mae])
        self.log_metric("accuracy", str(acc.get()[1]))
        self.log_metric("mae", str(mae.get()[1]))

        with self.report.open("w") as fp:
            fp.write(str(acc) + "\n")
            fp.write(str(mae))


class PredictDigits(PipelineTask):
    batch_size = parameter[int].default(100)
    model = output
    report = output

    def band(self):
        train_data = DownloadFile.from_web(digit_data.train_data).data
        train_labels = DownloadFile.from_web(digit_data.train_label).data
        test_data = DownloadFile.from_web(digit_data.test_data).data
        test_labels = DownloadFile.from_web(digit_data.test_label).data

        model_task = TrainModel(
            training_data=train_data,
            training_labels=train_labels,
            test_data=test_data,
            test_labels=test_labels,
        )
        self.report = ValidateModel(
            model=model_task.model,
            validation_data=test_data,
            validation_label=test_labels,
        )

        self.model = model_task.model



# © Copyright Databand.ai, an IBM Company 2022

import logging

import pandas as pd
import pytest

from dbnd_examples_orchestration.orchestration.features import (
    example_dict_of_data_frames,
    example_folder,
    example_output_per_id,
)

from dbnd.testing.orchestration_utils import TargetTestBase
from dbnd_run.testing.helpers import assert_run_task
from targets.dir_target import DirTarget


logger = logging.getLogger(__name__)


@pytest.fixture
def sample():
    return pd.DataFrame(data=[[1, 1], [2, 2]], columns=["c1", "c2"])


class TestFeatureOutputsExamples(TargetTestBase):
    def test_dict_of_data(self, sample):
        task = example_dict_of_data_frames.create_multiple_data_frames.task(sample)
        task = assert_run_task(task)
        assert task
        assert isinstance(task.result, DirTarget)
        logger.info("All partitions %s" % task.result.list_partitions())
        assert len(task.result.list_partitions()) == 3

    def test_tutorial_partners_report(self, sample):
        task = example_folder.plot_graphs.task(data=sample)
        task = assert_run_task(task)
        assert task

    def test_tutorial_output_per_id(self, sample):
        task = example_output_per_id.output_per_id_report.task(partners=[1, 2])
        task = assert_run_task(task)
        assert task

    def test_folder_status(self):
        dir = self.target("dir.csv/")
        dir.mark_success()
        task = example_folder.read_folder_status.task(folder_input=dir.path)
        task = assert_run_task(task)
        assert task

    def test_folder_status_fails(self):
        dir = self.target("dir.csv/")
        with pytest.raises(Exception):
            task = example_folder.read_folder_status.task(folder_input=dir.path)
            assert_run_task(task)



# © Copyright Databand.ai, an IBM Company 2022

"""
Override build in implementation of Hdf5 serialization
"""
import pandas as pd

from dbnd import output, task
from targets.marshalling import DataFrameToHdf5, register_marshaller
from targets.target_config import FileFormat


class DataFrameToHdf5Table(DataFrameToHdf5):
    def _pd_to(self, value, *args, **kwargs):
        # WE WILL CHANGE THE DEFAULT FORMAT FOR MARSHALLER
        kwargs.setdefault("format", "table")
        return super(DataFrameToHdf5Table, self)._pd_to(value, *args, **kwargs)


register_marshaller(pd.DataFrame, FileFormat.hdf5, DataFrameToHdf5Table())


@task(result=output.hdf5)
def dump_as_hdf5_table():
    # type: ()-> pd.DataFrame
    return pd.DataFrame(
        data=list(zip(["Bob", "Jessica"], [968, 155])), columns=["Names", "Births"]
    )



# © Copyright Databand.ai, an IBM Company 2022

"""import pyspark.sql as spark

from dbnd import PipelineTask, output, parameter
from dbnd_examples_orchestration.dbnd_spark import spark_folder
from dbnd_spark import SparkConfig
from dbnd_spark.spark import PySparkTask, SparkTask, spark_task


#### DOC START


@spark_task(result=output[spark.DataFrame])
def word_count_inline(text=parameter.txt[spark.DataFrame]):
    from operator import add
    from dbnd_spark.spark import get_spark_session

    lines = text.rdd.map(lambda r: r[0])
    counts = (
        lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
    )

    return get_spark_session().createDataFrame(counts.collect())


class WordCountPySparkTask(PySparkTask):
    text = parameter.data
    counters = output

    python_script = spark_folder("pyspark/word_count.py")

    def application_args(self):
        return [self.text, self.counters]


class WordCountTask(SparkTask):
    text = parameter.data
    counters = output

    main_class = "ai.databand.examples.WordCount"
    defaults = {
        SparkConfig.main_jar: "${DBND_HOME}/databand_examples/tool_spark/spark_jvm/target/ai.databand.examples-1.0-SNAPSHOT.jar"
    }

    def application_args(self):
        return [self.text, self.counters]


class AllWordCounts(PipelineTask):
    text = parameter.data

    counters_python = output.csv.data
    counters_java = output.csv.data

    def band(self):
        t_text = CloneFile(text=self.text)

        self.counters_python = WordCountPySparkTask(text=t_text.clone)
        self.counters_java = WordCountTask(t_text.clone)


#### DOC END"""



# © Copyright Databand.ai, an IBM Company 2022

__version__ = "0.1.0"

__title__ = "dbnd-examples"
__description__ = "databand.ai example project"
__url__ = "http://www.databand.ai/"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Evgeny Shulman"
__email__ = "evgeny.shulman@databand.ai"

__license__ = "Commercial Licence"
__copyright__ = "Copyright (c) 2018 databand.ai"



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os

from typing import List

from airflow.hooks.S3_hook import S3Hook
from dbnd_examples_orchestration.orchestration.dbnd_aws.sync_operators import S3SyncHook

from dbnd import output, parameter, pipeline, task
from dbnd_aws.batch.aws_batch_ctrl import AwsBatchConfig
from dbnd_docker.docker.docker_task import DockerRunTask
from targets import target
from targets.fs import FileSystems
from targets.types import PathStr
from targets.utils.path import path_to_bucket_and_key


logger = logging.getLogger(__name__)


@task
def copy_to_s3_destination(source: List[str], destination: PathStr = output[PathStr]):
    hook = S3Hook()

    for source_file in source:
        _, source_key = path_to_bucket_and_key(source_file)
        destination_file = os.path.join(destination, source_key)

        logger.info("Copying %s -> %s", source_file, destination_file)
        if target(source_file).fs.name == FileSystems.s3:

            hook.copy_object(
                source_bucket_key=source_file, dest_bucket_key=destination_file
            )
        else:

            hook.load_file(filename=source_file, key=destination_file, replace=True)
    return source


class ValidatePartnerData(DockerRunTask):
    input_data = parameter[PathStr]
    output_validated = parameter.output[PathStr]
    command = "echo {{ params.input_data}} {{params.output_validated}} "
    image = "ubuntu:latest"

    # DockerRunTask can run on as local docker container and as  AwsBatch job.
    # Next lines define configuration parameters for AwsBatch controller
    defaults = {
        AwsBatchConfig.job_definition: "hello-world",
        AwsBatchConfig.job_queue: "first-run-job-queue",
    }

    def _task_submit(self):
        # override task submit method to add custom report after task (docker command) was executed
        v = super(ValidatePartnerData, self)._task_submit()

        self.output_validated.write("aaa")
        return v


@pipeline(destination=output)
def partner_data_ingest(
    new_files: List[str],
    partner_id=1,
    destination: PathStr = "s3://databand-victor-test/destination",
):
    """
    :param new_files : list of files to process:
    :param partner_id:
    :param destination: destination to copy validated files
    This pipeline gets list of files and run ValidatePartnerData on an every one of them, all validated files are
    then copised into a destination folder.
    """

    validated_reports = [ValidatePartnerData(input_data=f) for f in new_files]
    logger.info("%s ", validated_reports)
    published_files = copy_to_s3_destination(source=new_files, destination=destination)

    # set all validated_reports tasks as upstream for copy_to_s3_destination.
    # this would make copy_to_s3_destination only after all validated_reports finished successfully
    for vr in validated_reports:
        published_files.set_upstream(vr)

    # collect outputs (i.e. validation report) of validation tasks into a list
    reports = [vr.output_validated for vr in validated_reports]
    return published_files.result, reports


@pipeline(destination=output)
def partner_data_ingest_new_files(source, destination):
    """
    :param source : list of files to process:
    :param destination: destination to copy validated files
    check s3 path for new file, trigger partner_data_ingest for new files.
    """
    hook = S3SyncHook(aws_conn_id="aws_default", verify=True)

    diff = hook.diff(source, destination)
    return partner_data_ingest(new_files=diff, destination=destination)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd import pipeline
from dbnd_aws.sagemaker import SageMakerTrainTask
from dbnd_aws.sagemaker.estimator_config import PyTorchEstimatorConfig


@pipeline
def train_mnist_with_pytorch(train: str):
    hyperparameters = {"epochs": 6, "backend": "gloo"}

    train = SageMakerTrainTask(
        train=train,
        estimator_config=PyTorchEstimatorConfig(
            entry_point="dbnd_examples_orchestration/tool_sagemaker/mnist.py",
            framework_version="1.1.0",
            hyperparameters=hyperparameters,
        ),
    )
    return train.output_path



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd import dbnd_config, get_databand_context, pipeline, task
from dbnd._core.context.databand_context import DatabandContext


@task
def say_hello(user="sdfsd"):
    greeting = "Hey, {}!".format(user)
    logging.info(greeting)
    return greeting


@task
def calculate_num_of_greetings(input=2):
    return input * 2


@pipeline
def greetings_pipeline(num_of_greetings):
    results = []
    for i in range(num_of_greetings):
        v = say_hello(user="user {}".format(i))
        results.append(v)
    return results


@task
def greetings_pipeline_subrun(num_of_greetings):
    # this is task, so num_of_greetings is going to be evaluated
    assert isinstance(num_of_greetings, int)

    dc = get_databand_context()  # type: DatabandContext
    # if you are running in "submission" mode, you should cancel that for the new run.
    dc.settings.run.submit_driver = False

    greetings_pipeline.task(num_of_greetings=num_of_greetings).dbnd_run()
    return "OK"


@pipeline
def say_hello_pipe():
    num_of_greetings = calculate_num_of_greetings()
    # we can't use num_of_greetigs value here - we are inside pipeline build phase
    v = greetings_pipeline_subrun(num_of_greetings=num_of_greetings)

    return v


if __name__ == "__main__":
    with dbnd_config(
        {
            say_hello_pipe.task.task_version: "now",
            calculate_num_of_greetings.task.task_version: "1",
        }
    ):
        say_hello_pipe.dbnd_run()



# © Copyright Databand.ai, an IBM Company 2022

from dbnd import task


class TestDocRunningPipelines:
    def test_doc(self):
        #### DOC START

        @task
        def prepare_data(data: str) -> str:
            return data

        #### DOC END
        prepare_data.dbnd_run(data="Hello Databand!")



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Dict

from pandas import DataFrame

from dbnd import output, task


logger = logging.getLogger(__name__)


@task(result=output.folder[Dict[str, DataFrame]])
def create_multiple_data_frames(data):
    # type: (DataFrame) -> Dict[str, DataFrame]

    return {"a": data.head(1), "b": data.head(2), "c": data.head(1)}



# © Copyright Databand.ai, an IBM Company 2022

### ORIGINAL CODE
import logging

from typing import Tuple

import numpy as np
import pandas as pd

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import train_test_split

from dbnd import config, log_dataframe, log_metric, pipeline, task
from dbnd_run.testing.helpers import assert_run_task


logger = logging.getLogger(__name__)


def my_training_script():
    # load data
    raw_data = pd.read_csv(data_repo.wines_full)

    # Originally in the documentation we have this line instead:
    # raw_data = pd.read_csv("wine.csv")

    # split data into training and validation sets
    train_df, validation_df = train_test_split(raw_data, test_size=0.5)

    # create hyperparameters and model
    alpha = 0.5
    l1_ratio = 0.2
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
    lr.fit(train_df.drop(["quality"], 1), train_df[["quality"]])

    # validation
    validation_x = validation_df.drop(["quality"], 1)
    validation_y = validation_df[["quality"]]
    prediction = lr.predict(validation_x)
    rmse = np.sqrt(mean_squared_error(validation_y, prediction))
    mae = mean_absolute_error(validation_y, prediction)
    r2 = r2_score(validation_y, prediction)

    logging.info("%s,%s,%s", rmse, mae, r2)

    return lr


@task(result="training_set, validation_set")
def prepare_data(raw_data: DataFrame) -> Tuple[DataFrame, DataFrame]:
    train_df, validation_df = train_test_split(raw_data)

    return train_df, validation_df


@task
def train_model(
    training_set: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
) -> ElasticNet:
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
    lr.fit(training_set.drop(["quality"], 1), training_set[["quality"]])

    return lr


@task
def validate_model(model: ElasticNet, validation_dataset: DataFrame) -> str:
    log_dataframe("validation", validation_dataset)

    validation_x = validation_dataset.drop(["quality"], 1)
    validation_y = validation_dataset[["quality"]]

    prediction = model.predict(validation_x)
    rmse = np.sqrt(mean_squared_error(validation_y, prediction))
    mae = mean_absolute_error(validation_y, prediction)
    r2 = r2_score(validation_y, prediction)

    log_metric("rmse", rmse)
    log_metric("mae", rmse)
    log_metric("r2", r2)

    return "%s,%s,%s" % (rmse, mae, r2)


@pipeline(result=("model", "validation"))
def predict_wine_quality(
    raw_data: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
):
    training_set, validation_set = prepare_data(raw_data=raw_data)

    model = train_model(training_set=training_set, alpha=alpha, l1_ratio=l1_ratio)
    validation = validate_model(model=model, validation_dataset=validation_set)

    return model, validation


class TestQuickStartExample(object):
    def test_predict_wine_quality(self):
        task_config = {"raw_data": data_repo.wines_full}
        config.set_values(
            {
                predict_wine_quality.task_cls.task_definition.task_config_section: task_config
            },
            source="--set-root",
        )
        assert_run_task(predict_wine_quality.task())

    def test_script(self):
        result = my_training_script()
        assert result is not None



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd_examples_orchestration.orchestration.tutorial_syntax.T12_data_pandas_and_numpy import (
    f_test_pandas_numpy_flow,
)

from dbnd_run.testing.helpers import assert_run_task


logger = logging.getLogger(__name__)


class TestTutorialSyntax(object):
    def test_pandas_numpy(self):
        assert_run_task(f_test_pandas_numpy_flow.task())



# © Copyright Databand.ai, an IBM Company 2022

"""from dbnd import config, output, parameter
from dbnd_spark import spark, spark_task
from dbnd_spark.local.local_spark_config import SparkLocalEngineConfig

def assert_run_task(task):  # type: (Union[T, Task]) -> Union[T, Task]
    task.dbnd_run()
    assert task._complete()
    return task

#### DOC START


@spark_task(result=output[spark.DataFrame])
def word_count_inline(data=parameter.csv[spark.DataFrame]):
    # spark business logic goes here
    # set a breakpoint here with a debuger of your choice
    ...


# invoke spark task  this way
if __name__ == "__main__":
    # create spark context and run spark task inside this context
    with spark.SparkSession.builder.getOrCreate() as sc:
        word_count_inline.dbnd_run(text=__file__)


def test_spark_inline_same_context(self):
    from pyspark.sql import SparkSession
    from dbnd_examples_orchestration.orchestration.dbnd_spark.word_count import word_count_inline

    with SparkSession.builder.getOrCreate() as sc:
        with config({SparkLocalEngineConfig.enable_spark_context_inplace: True}):
            task_instance = word_count_inline.t(text=__file__)
            assert_run_task(task_instance)


#### DOC END"""



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Tuple

import numpy as np
import pandas as pd

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score

from dbnd import log_dataframe, log_metric, pipeline, task


def train_test_split(raw_data):
    return raw_data, raw_data


class TestDocOrchestrationQuickstart:
    def test_calculate_alpha(self):
        #### DOC START
        from dbnd import task

        @task
        def calculate_alpha(alpha: float = 0.5) -> float:
            alpha += 0.1
            return alpha

        #### DOC END
        calculate_alpha.dbnd_run(alpha=0.4)

    def test_without_dbnd(self):
        #### DOC START
        logging.basicConfig(level=logging.INFO)

        def training_script():
            # load data
            raw_data = pd.read_csv(data_repo.wines)

            # split data into training and validation sets
            train_df, validation_df = train_test_split(raw_data)

            # create hyperparameters and model
            alpha = 0.5
            l1_ratio = 0.2
            lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
            lr.fit(train_df.drop(["quality"], 1), train_df[["quality"]])

            # validation
            validation_x = validation_df.drop(["quality"], 1)
            validation_y = validation_df[["quality"]]
            prediction = lr.predict(validation_x)
            rmse = np.sqrt(mean_squared_error(validation_y, prediction))
            mae = mean_absolute_error(validation_y, prediction)
            r2 = r2_score(validation_y, prediction)

            logging.info("%s,%s,%s", rmse, mae, r2)

            return lr

        training_script()
        #### DOC END

    def test_using_dbnd(self):
        logging.basicConfig(level=logging.INFO)

        @task(result="training_set, validation_set")
        def prepare_data(raw_data: DataFrame) -> Tuple[DataFrame, DataFrame]:
            train_df, validation_df = train_test_split(raw_data)

            return train_df, validation_df

        @task
        def train_model(
            training_set: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
        ) -> ElasticNet:
            lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
            lr.fit(training_set.drop(["quality"], 1), training_set[["quality"]])

            return lr

        @task
        def validate_model(model: ElasticNet, validation_dataset: DataFrame) -> str:
            log_dataframe("validation", validation_dataset)

            validation_x = validation_dataset.drop(["quality"], 1)
            validation_y = validation_dataset[["quality"]]

            prediction = model.predict(validation_x)
            rmse = np.sqrt(mean_squared_error(validation_y, prediction))
            mae = mean_absolute_error(validation_y, prediction)
            r2 = r2_score(validation_y, prediction)

            log_metric("rmse", rmse)
            log_metric("mae", rmse)
            log_metric("r2", r2)

            return "%s,%s,%s" % (rmse, mae, r2)

        @pipeline(result=("model", "validation"))
        def predict_wine_quality(
            raw_data: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
        ):
            training_set, validation_set = prepare_data(raw_data=raw_data)

            model = train_model(
                training_set=training_set, alpha=alpha, l1_ratio=l1_ratio
            )
            validation = validate_model(model=model, validation_dataset=validation_set)

            return model, validation

        #### DOC END
        predict_wine_quality.dbnd_run(raw_data=data_repo.wines)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.data import dbnd_examples_data_path

from dbnd_gcp.apache_beam.apache_beam_task import ApacheBeamPythonTask
from dbnd_gcp.apache_beam.parameters import beam_input, beam_output


class BeamWordCount(ApacheBeamPythonTask):
    py_file = dbnd_examples_data_path("dbnd_gcp/tool_dataflow/pybeam/wordcount.py")

    input = beam_input
    output = beam_output

    def get_beam_task_options(self):
        self.output.mkdir()
        return {"input": self.input, "output": self.output}



# © Copyright Databand.ai, an IBM Company 2022

from typing import Tuple

import pandas as pd

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame
from sklearn.linear_model import ElasticNet
from sklearn.model_selection import train_test_split

from dbnd import pipeline, task


#### DOC START


# define a function with a decorator


@task(result="training_set, validation_set")
def prepare_data(raw_data: DataFrame) -> Tuple[DataFrame, DataFrame]:
    train_df, validation_df = train_test_split(raw_data)

    return train_df, validation_df


@task
def train_model(
    training_set: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
) -> ElasticNet:
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
    lr.fit(training_set.drop(["quality"], 1), training_set[["quality"]])

    return lr


@pipeline(result=("model", "validation_set"))
def predict_wine_quality(
    raw_data: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
):
    training_set, validation_set = prepare_data(raw_data=raw_data)

    model = train_model(training_set=training_set, alpha=alpha, l1_ratio=l1_ratio)

    return model, validation_set


#### DOC END


class TestDocOrchestratingTasksAndPipelines:
    def test_doc(self):
        predict_wine_quality.task(raw_data=pd.read_csv(data_repo.wines)).dbnd_run()



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import List

import pandas as pd

from dbnd import config, dbnd_handle_errors, pipeline, task


logger = logging.getLogger(__name__)


@task
def operation_A(x_input="x"):
    logger.info("Running  %s -> operation_x", x_input)
    if x_input == "ha":
        raise Exception()
    return "{} -> operation_x".format(x_input)


@pipeline
def pipe_A_operations(pipe_argument="pipe"):
    z = operation_A(pipe_argument)
    x = operation_A(z)
    y = operation_A(x, task_name="zzz")

    # this operation is not wired to any outputs or return values
    # but we need it to run, so it will be "related" to pipe_operations automatically
    operation_A("standalone")

    # you will see y outputs as pipe_argument output in UI
    return y


if __name__ == "__main__":
    with config({"zzz": {"x_input": "ha2"}}):
        operations_task = pipe_A_operations.task(task_version="now")
        operations_task.dbnd_run()


@task
def get_input_data() -> pd.DataFrame:
    return pd.DataFrame([["A"], ["B"], ["C"]], columns=["Names"])


@task
def run_experiment(df: pd.DataFrame, exp_name, ratio) -> pd.DataFrame:
    return pd.DataFrame(
        [[exp_name, ratio, len(df)]], columns=["Names", "Ratio", "Length"]
    )


@task
def build_experiment_report(exp_results: List[pd.DataFrame]) -> pd.DataFrame:
    return pd.concat(exp_results)


@pipeline
def experiment_pipeline():
    input_data = get_input_data()
    experiments = [
        run_experiment(input_data, "exp_%s" % ratio, ratio) for ratio in range(100)
    ]
    report = build_experiment_report(experiments)
    return report


@task
def gen_token(t):
    return t


@task
def concat_tokens(tokens_list: List[str]):
    logging.warning("TOKENS: %s", tokens_list)


@pipeline
def experiment_concat():
    # if type is specified, we can mix between strings and task outputs
    l = ["aa", gen_token("bbb")]
    return concat_tokens(l)

    # this operation is not wired to any outputs or return values
    # but we need it to run, so it will be "related" to pipe_operations automatically
    operation_z("standalone")


if __name__ == "__main__":
    with dbnd_handle_errors():
        experiment_pipeline.dbnd_run(task_version="now")



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.orchestration.tool_mxnet import (
    mxnet_digit_mnist,
    mxnet_fashion_mnist,
)


__all__ = ["mxnet_digit_mnist", "mxnet_fashion_mnist"]



# © Copyright Databand.ai, an IBM Company 2022

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# This file has been modified by databand.ai to include more advance example.

import sys

from operator import add

from pyspark.sql import SparkSession

from dbnd import log_dataframe, task


@task
def word_count(input_path, output_path):
    spark = SparkSession.builder.appName("PythonWordCount").getOrCreate()
    lines = spark.read.text(input_path)

    log_dataframe("lines", lines)
    lines = lines.rdd.map(lambda r: r[0])

    log_dataframe("lines_rdd", lines)
    counts = (
        lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
    )
    counts.saveAsTextFile(output_path)
    output = counts.collect()
    log_dataframe("output", output)
    for (word, count) in output:
        print("%s: %i" % (word, count))
    # this makes trouble on job submit on databricks!
    # spark.close()


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: wordcount <file> <output>")
        sys.exit(-1)

    word_count(sys.argv[1], sys.argv[2])



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Dict

import attr

from dbnd import band, task
from targets.values import ValueType, register_value_type


logger = logging.getLogger(__name__)


###############
# USER PART
@attr.s
class MyConfig(object):
    my_config_id = attr.ib()
    data = attr.ib()


#######################
# DATABAND INTEGRATION
class MyConfigValueType(ValueType):
    type = MyConfig

    def parse_from_str(self, x):
        # calculate config
        logger.warning("Calculating my_config with id %s", x)
        return MyConfig(my_config_id=x, data={"config": "1"})


register_value_type(MyConfigValueType())


###############
# USAGE


@task
def update_config(my_config=1):
    # type: (MyConfig)-> MyConfig
    new_config = MyConfig(
        my_config_id="{}_patched".format(my_config.my_config_id), data={"config": "2"}
    )

    return new_config


@task
def use_config(my_config):
    # type: (MyConfig)-> Dict
    assert my_config.data["config"] == "2"
    return my_config.data


@band(result=("updated", "config_data"))
def config_flow(my_config):
    # type: (MyConfig)->  (MyConfig, Dict)
    updated_config = update_config(my_config)
    config_data = use_config(updated_config)
    return updated_config, config_data



# © Copyright Databand.ai, an IBM Company 2022


class SagemakerData(object):
    amazon_reviews_raw_data = "s3://amazon-reviews-pds/tsv/amazon_reviews_us_Digital_Video_Download_v1_00.tsv.gz"


sagemaker_data_repo = SagemakerData()



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd_examples_orchestration.orchestration.examples.salad import prepare_salad

from dbnd_run.testing.helpers import assert_run_task


logger = logging.getLogger(__name__)


class TestTutorialSyntax(object):
    def test_prepare_salad(self):
        assert_run_task(prepare_salad.task())



# © Copyright Databand.ai, an IBM Company 2022

"""
This is a way to add another format to serialize/deserialize SizedMessage example object
"""
import logging

import joblib

from dbnd import output, task
from targets.marshalling import register_marshaller
from targets.marshalling.marshaller import Marshaller
from targets.target_config import TargetConfig, register_file_extension
from targets.values import ValueType, register_value_type


logger = logging.getLogger(__name__)


class SizedMessage(object):
    def __init__(self, msg, size):
        self.msg = msg
        self.size = size


# This defines how SizedMessage is parsed and serialised into strings
class MessageValueType(ValueType):
    type = SizedMessage

    def parse_from_str(self, x):
        parts = x.split(">")
        return SizedMessage(parts[0], parts[1])

    def to_str(self, x):
        return x.msg + ">" + str(x.size)


# This registers value type with value type registry
register_value_type(MessageValueType())


# 1. create file extension
z_file_ext = register_file_extension("z")


class JoblibSizedMessageMarshaller(Marshaller):
    def target_to_value(self, target, **kwargs):
        with target.open() as fp:
            from_file = joblib.load(fp.name)
            return from_file

    def value_to_target(self, value, target, **kwargs):
        with target.open("w") as fp:
            joblib.dump(value, fp.name)


# 2. register type to extension mapping
register_marshaller(SizedMessage, z_file_ext, JoblibSizedMessageMarshaller())


@task(result=output.target_config(TargetConfig(format=z_file_ext)))
def dump_as_joblib():
    # type: ()-> SizedMessage
    return SizedMessage("example message \n", 10)


@task(result=output.txt[int])
def load_as_joblib(sized_message: SizedMessage):
    return sized_message.msg * sized_message.size



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging

from pathlib import Path

from dbnd_examples_orchestration.orchestration.tutorial_syntax.T02_task_parameters import (
    f_path_types,
    f_task_parameters,
)

from dbnd import pipeline
from targets import target


logger = logging.getLogger(__name__)

EXPECTED_STR = (
    "'1' 'False' 'strstr'",
    "'2018-01-01 00:00:00' '2018-01-01' '1 day, 0:00:00'",
)

some_existing_file = __file__


@pipeline
def f_run_all_simple():
    result = f_task_parameters(
        1,
        False,
        "strstr",
        v_datetime=datetime.datetime(year=2018, month=1, day=1),
        v_date=datetime.date(year=2018, month=1, day=1),
        v_period=datetime.timedelta(days=1),
        v_list=[1, "1"],
        v_list_obj=[2, "2"],
        v_list_str=["str", "str"],
        v_list_int=[1, 2],
        v_list_date=[datetime.date(year=2018, month=1, day=1)],
    )
    paths = f_path_types(
        pathlib_path=Path(some_existing_file),
        str_as_path=some_existing_file,
        target_path=target(some_existing_file),
    )

    return result, paths


class TestTaskParameters(object):
    def test_run_all_as_regular_function(self):
        result, paths = f_run_all_simple()
        # this was a normal execution of the function, no databand required!

        assert result

    def test_run_all_task(self):
        # the test code runs outsider @band context, we need to explicitly state .task() for a function
        # otherwise it will be just called like a regular function

        run_result = f_run_all_simple.task().dbnd_run()

        actual = run_result.root_task.result[0].load(object)
        assert actual



# © Copyright Databand.ai, an IBM Company 2022

import pandas as pd

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame

from dbnd import pipeline, task


class TestDocTasksPipelinesData:
    def test_pipelines(self):
        #### DOC START
        @task
        def prepare_data(data: pd.DataFrame) -> pd.DataFrame:
            return data

        @task
        def train_model(data: pd.DataFrame) -> object:
            ...

        @pipeline
        def prepare_data_pipeline(data: pd.DataFrame):
            prepared_data = prepare_data(data)
            model = train_model(prepared_data)
            return model

        #### DOC END
        prepare_data_pipeline.dbnd_run(data_repo.wines)

    def test_data(self):
        #### DOC START
        @task
        def prepare_data(data: DataFrame) -> DataFrame:
            data["new_column"] = 5
            return data

        #### DOC END
        prepare_data.dbnd_run(data=data_repo.wines)



# © Copyright Databand.ai, an IBM Company 2022

# Based on https://github.com/awslabs/amazon-sagemaker-examples/blob/master/sagemaker-python-sdk/pytorch_mnist/mnist.py
# Distriburted under Apache License 2.0
# Converted into databand example

import argparse
import json
import logging
import os
import sys

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data
import torch.utils.data.distributed

from torchvision import datasets, transforms


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler(sys.stdout))


# Based on https://github.com/pytorch/examples/blob/master/mnist/main.py
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


def _get_train_data_loader(batch_size, training_dir, is_distributed, **kwargs):
    logger.info("Get train data loader")
    dataset = datasets.MNIST(
        training_dir,
        train=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
        download=True,
    )
    train_sampler = (
        torch.utils.data.distributed.DistributedSampler(dataset)
        if is_distributed
        else None
    )
    return torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=train_sampler is None,
        sampler=train_sampler,
        **kwargs
    )


def _get_test_data_loader(test_batch_size, training_dir, **kwargs):
    logger.info("Get test data loader")
    return torch.utils.data.DataLoader(
        datasets.MNIST(
            training_dir,
            train=False,
            transform=transforms.Compose(
                [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
            ),
            download=True,
        ),
        batch_size=test_batch_size,
        shuffle=True,
        **kwargs
    )


def _average_gradients(model):
    # Gradient averaging.
    size = float(dist.get_world_size())
    for param in model.parameters():
        dist.all_reduce(param.grad.data, op=dist.reduce_op.SUM)
        param.grad.data /= size


def train(args):
    is_distributed = len(args.hosts) > 1 and args.backend is not None
    logger.debug("Distributed training - {}".format(is_distributed))
    use_cuda = args.num_gpus > 0
    logger.debug("Number of gpus available - {}".format(args.num_gpus))
    kwargs = {"num_workers": 1, "pin_memory": True} if use_cuda else {}
    device = torch.device("cuda" if use_cuda else "cpu")

    if is_distributed:
        # Initialize the distributed environment.
        world_size = len(args.hosts)
        os.environ["WORLD_SIZE"] = str(world_size)
        host_rank = args.hosts.index(args.current_host)
        os.environ["RANK"] = str(host_rank)
        dist.init_process_group(
            backend=args.backend, rank=host_rank, world_size=world_size
        )
        logger.info(
            "Initialized the distributed environment: '{}' backend on {} nodes. ".format(
                args.backend, dist.get_world_size()
            )
            + "Current host rank is {}. Number of gpus: {}".format(
                dist.get_rank(), args.num_gpus
            )
        )

    # set the seed for generating random numbers
    torch.manual_seed(args.seed)
    if use_cuda:
        torch.cuda.manual_seed(args.seed)

    train_loader = _get_train_data_loader(
        args.batch_size, args.data_dir, is_distributed, **kwargs
    )
    test_loader = _get_test_data_loader(args.test_batch_size, args.data_dir, **kwargs)

    logger.debug(
        "Processes {}/{} ({:.0f}%) of train data".format(
            len(train_loader.sampler),
            len(train_loader.dataset),
            100.0 * len(train_loader.sampler) / len(train_loader.dataset),
        )
    )

    logger.debug(
        "Processes {}/{} ({:.0f}%) of test data".format(
            len(test_loader.sampler),
            len(test_loader.dataset),
            100.0 * len(test_loader.sampler) / len(test_loader.dataset),
        )
    )

    model = Net().to(device)
    if is_distributed and use_cuda:
        # multi-machine multi-gpu case
        model = torch.nn.parallel.DistributedDataParallel(model)
    else:
        # single-machine multi-gpu case or single-machine or multi-machine cpu case
        model = torch.nn.DataParallel(model)

    optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    for epoch in range(1, args.epochs + 1):
        model.train()
        for batch_idx, (data, target) in enumerate(train_loader, 1):
            data, target = data.to(device), target.to(device)
            optimizer.zero_grad()
            output = model(data)
            loss = F.nll_loss(output, target)
            loss.backward()
            if is_distributed and not use_cuda:
                # average gradients manually for multi-machine cpu case only
                _average_gradients(model)
            optimizer.step()
            if batch_idx % args.log_interval == 0:
                logger.info(
                    "Train Epoch: {} [{}/{} ({:.0f}%)] Loss: {:.6f}".format(
                        epoch,
                        batch_idx * len(data),
                        len(train_loader.sampler),
                        100.0 * batch_idx / len(train_loader),
                        loss.item(),
                    )
                )
        test(model, test_loader, device)
    save_model(model, args.model_dir)


def test(model, test_loader, device):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(
                output, target, size_average=False
            ).item()  # sum up batch loss
            pred = output.max(1, keepdim=True)[
                1
            ]  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)
    logger.info(
        "Test set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n".format(
            test_loss,
            correct,
            len(test_loader.dataset),
            100.0 * correct / len(test_loader.dataset),
        )
    )


def model_fn(model_dir):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = torch.nn.DataParallel(Net())
    with open(os.path.join(model_dir, "model.pth"), "rb") as f:
        model.load_state_dict(torch.load(f))
    return model.to(device)


def save_model(model, model_dir):
    logger.info("Saving the model.")
    path = os.path.join(model_dir, "model.pth")
    # recommended way from http://pytorch.org/docs/master/notes/serialization.html
    torch.save(model.cpu().state_dict(), path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # Data and model checkpoints directories
    parser.add_argument(
        "--batch-size",
        type=int,
        default=64,
        metavar="N",
        help="input batch size for training (default: 64)",
    )
    parser.add_argument(
        "--test-batch-size",
        type=int,
        default=1000,
        metavar="N",
        help="input batch size for testing (default: 1000)",
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=10,
        metavar="N",
        help="number of epochs to train (default: 10)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=0.01,
        metavar="LR",
        help="learning rate (default: 0.01)",
    )
    parser.add_argument(
        "--momentum",
        type=float,
        default=0.5,
        metavar="M",
        help="SGD momentum (default: 0.5)",
    )
    parser.add_argument(
        "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
    )
    parser.add_argument(
        "--log-interval",
        type=int,
        default=100,
        metavar="N",
        help="how many batches to wait before logging training status",
    )
    parser.add_argument(
        "--backend",
        type=str,
        default=None,
        help="backend for distributed training (tcp, gloo on cpu and gloo, nccl on gpu)",
    )

    # Container environment
    parser.add_argument(
        "--hosts", type=list, default=json.loads(os.environ["SM_HOSTS"])
    )
    parser.add_argument(
        "--current-host", type=str, default=os.environ["SM_CURRENT_HOST"]
    )
    parser.add_argument("--model-dir", type=str, default=os.environ["SM_MODEL_DIR"])
    parser.add_argument(
        "--data-dir", type=str, default=os.environ["SM_CHANNEL_TRAINING"]
    )
    parser.add_argument("--num-gpus", type=int, default=os.environ["SM_NUM_GPUS"])

    train(parser.parse_args())



# © Copyright Databand.ai, an IBM Company 2022

from os import path

from setuptools import setup
from setuptools.config import read_configuration


BASE_PATH = path.join(path.dirname(path.abspath(__file__)))
CFG_PATH = path.join(BASE_PATH, "setup.cfg")
config = read_configuration(CFG_PATH)

version = config["metadata"]["version"]


INSTALL_REQUIRES = [
    # we are still installing 'databand' in dockers.. "dbnd==" + version,
    "dbnd==" + version,
    'scikit-learn==0.23.2;python_version<"3.8"',
    'scikit-learn==1.2.0;python_version>="3.8"',
    'scipy==1.1.0;python_version<"3.8"',
    'scipy==1.9.3;python_version>="3.8"',
    'matplotlib==3.3.0;python_version<"3.8"',
    'matplotlib==3.6.2;python_version>="3.8"',
    'pyspark==2.4.4;python_version<"3.8"',
    'pyspark==3.3.1;python_version>="3.8"',
    "tables==3.7.0",
    "pandas<2.0.0,>=0.17.1",
]

setup(
    name="dbnd-examples-orchestration",
    package_dir={"": "src"},
    version=version,
    zip_safe=False,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    entry_points={
        "dbnd": ["dbnd-examples-orchestration = dbnd_examples_orchestration._plugin"]
    },
)



# © Copyright Databand.ai, an IBM Company 2022

"""
# Run specific task, increment task version
dbnd run dbnd_examples_orchestration.orchestration.dbnd_spark.salad_spark.CutAtSpark --set  vegetables=README.md
dbnd run dbnd_examples_orchestration.orchestration.dbnd_spark.salad_spark.PrepareSaladAtSpark
"""

import logging

from dbnd_examples_orchestration.data import data_repo
from dbnd_examples_orchestration.orchestration.dbnd_spark.scripts import spark_script

from dbnd import data, output, parameter
from dbnd_run.tasks import PipelineTask
from dbnd_spark import PySparkTask


logger = logging.getLogger(__name__)


class CutAtSpark(PySparkTask):
    python_script = spark_script("cut_salad.py")

    vegetables = parameter.data
    chopped_vegetables = output

    def application_args(self):
        return [self.vegetables, self.chopped_vegetables]


class AddDressingAtSpark(PySparkTask):
    python_script = spark_script("add_dressing.py")

    dressing = parameter[str]
    chopped_vegetables = parameter.data
    salad = output.data

    def application_args(self):
        return [self.chopped_vegetables, self.dressing, self.salad]


class PrepareSaladAtSpark(PipelineTask):
    vegetables = data(default=data_repo.vegetables)
    dressing = parameter.value("oil")

    salad = output.data

    def band(self):
        s1 = CutAtSpark(vegetables=self.vegetables)
        self.salad = AddDressingAtSpark(
            chopped_vegetables=s1.chopped_vegetables, dressing=self.dressing
        )



# © Copyright Databand.ai, an IBM Company 2022

from pathlib import Path
from typing import NamedTuple, Tuple

import pandas as pd

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame

from dbnd import output, parameter, task
from targets.target_config import FileFormat


DEFAULT_OUTPUT_HDFS_PATH = data_repo.wines

Outputs1 = NamedTuple("Outputs1", training_set=DataFrame, real_data=DataFrame)


class TestDocOutputs:
    def test_prepare_data_named_output(self):
        #### DOC START
        @task(result="data")
        def prepare_data(data: DataFrame):
            return data

        #### DOC END
        prepare_data.dbnd_run(data=data_repo.wines)

    def test_prepare_data_txt_output(self):
        #### DOC START
        @task(result=output.txt)
        def prepare_data(data: DataFrame):
            return data

        #### DOC END

        prepare_data.dbnd_run(data=data_repo.wines)

    def test_prepare_data_defined_path(self):
        #### DOC START
        @task(custom_output=output[Path])
        def prepare_data(data_path: Path, custom_output: Path) -> str:
            with open(str(custom_output), "w") as fp:
                fp.write("my data successfully written")
            return str(data_path)

        #### DOC END
        # prepare_data.dbnd_run(
        #    data_path=data_repo.wines, custom_output="~/data/example.txt"
        # )

    def test_prepare_data_overwrite_previous_result(self):
        #### DOC START
        @task(
            result=output(default=DEFAULT_OUTPUT_HDFS_PATH).overwrite.csv[pd.DataFrame]
        )
        def prepare_data(data=parameter[pd.DataFrame]):
            return data

        #### DOC END
        prepare_data.dbnd_run(data=data_repo.wines)

    def test_prepare_data_tuples(self):
        #### DOC START
        @task(result="training_set,real_data")
        def prepare_data(data: DataFrame) -> Tuple[DataFrame, DataFrame]:
            data["new_column"] = 5
            return data, data

        #### DOC END
        prepare_data.dbnd_run(data_repo.wines)

    def test_prepare_data_two_outputs(self):
        #### DOC START
        @task(result=("training_set", "real_data"))
        def prepare_data(p: int = 1) -> (DataFrame, DataFrame):
            return (
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
            )

        #### DOC END
        prepare_data.dbnd_run()

    def test_prepare_data_no_output_names(self):
        #### DOC START
        @task
        def prepare_data(p: int) -> (DataFrame, DataFrame):
            return (
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
            )

        #### DOC END
        prepare_data.dbnd_run(p=1)

    def test_prepare_data_with_tuple(self):
        #### DOC START
        # Outputs = NamedTuple("outputs", training_set=DataFrame, real_data=DataFrame)

        @task
        def prepare_data(p: int) -> Outputs1:
            return Outputs1(
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
            )

        #### DOC END
        prepare_data.dbnd_run(p=1)

    def test_prepare_data_with_tuple_inside_task(self):
        #### DOC START
        @task
        def prepare_data(
            p: int,
        ) -> NamedTuple("Outputs", training_set=DataFrame, real_data=DataFrame):
            return (
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
            )

        #### DOC END
        prepare_data.dbnd_run(p=1)

    def test_no_tuple_multiple_outputs(self):
        #### DOC START
        @task(result="training_set,real_data")
        def prepare_data(p: int = 1) -> (DataFrame, DataFrame):
            return (
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
                pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
            )

        #### DOC END
        prepare_data.task(p=3).dbnd_run()

    def test_prepare_data_no_header(self):
        #### DOC START
        @task(result=parameter.csv.save_options(FileFormat.csv, header=False))
        def prepare_data(data: DataFrame) -> DataFrame:
            data["new_column"] = 5
            return data

        #### DOC END
        prepare_data.dbnd_run(data_repo.wines)



# © Copyright Databand.ai, an IBM Company 2022

import mxnet as mx

from dbnd_examples_orchestration.orchestration.tool_mxnet import (
    mnist_to_ndarray_iterator,
)
from mxnet.gluon import utils

from dbnd import PythonTask, output, parameter


class DownloadFile(PythonTask):
    task_target_date = None
    data_url = parameter[str]
    data_sha = parameter[str]

    data = output

    def run(self):
        utils.download(self.data_url, path=str(self.data), sha1_hash=self.data_sha)

    @classmethod
    def from_web(cls, file_url_and_sha1):
        return cls(data_url=file_url_and_sha1[0], data_sha=file_url_and_sha1[1])


class MXNetTask(PythonTask):
    seed = parameter.value(1)
    batch_size = parameter.value(100)

    def run(self):
        mx.random.seed(42)
        ctx = mx.gpu() if mx.test_utils.list_gpus() else mx.cpu()
        self.main(ctx=ctx)

    def to_ndarray_iterator(self, data_file, label_file, shuffle=False):
        return mnist_to_ndarray_iterator(
            data_file.path, label_file.path, self.batch_size, shuffle
        )


class MxNetGluonTask(MXNetTask):
    pass



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging
import time

from dbnd import parameter, pipeline, task


logger = logging.getLogger(__name__)


@task
class UserActionObject(object):
    def __init__(self, x_input="x", sleep_period=parameter[datetime.timedelta]):
        self.x_input = x_input
        self.sleep_period = sleep_period

    def run(self):
        logger.info(
            "Running  %s, %s -> UserActionObject", self.x_input, self.sleep_period
        )
        if self.x_input == "ha":
            raise Exception()
        if self.x_input == "sleep":
            time.sleep(self.sleep_period.total_seconds())
        return "{} -> operation_x".format(self.x_input)

    def on_kill(self):
        logger.error("Running on_kill at %s..", self)


@pipeline
def pipe_user_actions(pipe_argument="pipe"):
    step_1 = UserActionObject(x_input=pipe_argument)
    step_2 = UserActionObject(x_input=step_1)

    return step_2


if __name__ == "__main__":
    pipe_user_actions.dbnd_run()
